import enum

import cupy
import numpy as np
from karabo.bound import (
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
)

from . import base_gpu, base_calcat, utils
from ._version import version as deviceVersion
from .base_correction import BaseCorrection, add_correction_step_schema, preview_schema


class LpdConstants(enum.Enum):
    Offset = enum.auto()
    BadPixelsDark = enum.auto()
    GainAmpMap = enum.auto()
    FFMap = enum.auto()
    RelativeGain = enum.auto()
    BadPixelsFF = enum.auto()


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1
    GAIN_AMP = 2
    REL_GAIN = 4
    FF_CORR = 8
    BPMASK = 16


class LpdGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "lpd_gpu.cu"
    _corrected_axis_order = "cxy"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        input_data_dtype=cupy.uint16,
        output_data_dtype=cupy.float32,
        bad_pixel_mask_value=cupy.nan,
    ):
        self.input_shape = (memory_cells, 1, pixels_y, pixels_x)
        self.processed_shape = (memory_cells, pixels_y, pixels_x)
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            constant_memory_cells,
            input_data_dtype,
            output_data_dtype,
        )
        self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=cupy.float32)
        self.preview_buffer_getters.append(self._get_gain_map_for_preview)

        self.map_shape = (constant_memory_cells, pixels_x, pixels_y, 3)
        self.offset_map_gpu = cupy.zeros(self.map_shape, dtype=cupy.float32)
        self.gain_amp_map_gpu = cupy.ones(self.map_shape, dtype=cupy.float32)
        self.rel_gain_slopes_map_gpu = cupy.ones(self.map_shape, dtype=cupy.float32)
        self.flatfield_map_gpu = cupy.ones(self.map_shape, dtype=cupy.float32)
        self.bad_pixel_map_gpu = cupy.zeros(self.map_shape, dtype=cupy.uint32)
        self.bad_pixel_mask_value = bad_pixel_mask_value

        self.update_block_size((1, 1, 64))

    def _get_raw_for_preview(self):
        return self.input_data_gpu[:, 0]

    def _get_corrected_for_preview(self):
        return self.processed_data_gpu

    def _get_gain_map_for_preview(self):
        return self.gain_map_gpu

    def correct(self, flags):
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                cupy.uint8(flags),
                self.offset_map_gpu,
                self.gain_amp_map_gpu,
                self.rel_gain_slopes_map_gpu,
                self.flatfield_map_gpu,
                self.bad_pixel_map_gpu,
                self.bad_pixel_mask_value,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def load_constant(self, constant_type, constant_data):
        # constant type → transpose order
        bad_pixel_loading = {
            LpdConstants.BadPixelsDark: (2, 1, 0, 3),
            LpdConstants.BadPixelsFF: (2, 0, 1, 3),
        }
        # constant type → transpose order, GPU buffer
        other_constant_loading = {
            LpdConstants.Offset: ((2, 1, 0, 3), self.offset_map_gpu),
            LpdConstants.GainAmpMap: ((2, 0, 1, 3), self.gain_amp_map_gpu),
            LpdConstants.FFMap: ((2, 0, 1, 3), self.flatfield_map_gpu),
            LpdConstants.RelativeGain: ((2, 1, 0, 3), self.rel_gain_slopes_map_gpu),
        }
        if constant_type in bad_pixel_loading:
            self.bad_pixel_map_gpu |= cupy.asarray(
                np.transpose(
                    constant_data,
                    bad_pixel_loading[constant_type],
                ),
                dtype=np.uint32,
            )
        elif constant_type in other_constant_loading:
            transpose_order, gpu_buffer = other_constant_loading[constant_type]
            gpu_buffer.set(
                np.transpose(
                    constant_data.astype(np.float32),
                    transpose_order,
                )
            )
        else:
            raise ValueError(f"Unhandled constant type {constant_type}")

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")

    def flush_buffers(self):
        self.offset_map_gpu.fill(0)
        self.gain_amp_map_gpu.fill(1)
        self.rel_gain_slopes_map_gpu.fill(1)
        self.flatfield_map_gpu.fill(1)
        self.bad_pixel_map_gpu.fill(0)


class LpdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = LpdConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            LpdConstants.Offset: self.dark_condition,
            LpdConstants.BadPixelsDark: self.dark_condition,
            LpdConstants.GainAmpMap: self.category_condition,
            LpdConstants.FFMap: self.category_condition,
            LpdConstants.RelativeGain: self.category_condition,
            LpdConstants.BadPixelsFF: self.category_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        super(LpdCalcatFriend, LpdCalcatFriend).add_schema(
            schema, managed_keys, "LPD-Type", param_prefix, status_prefix
        )

        (
            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.memoryCells")
            .setNewDefaultValue(512)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key(f"{param_prefix}.biasVoltage")
            .setNewDefaultValue(250)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.feedbackCapacitor")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.3)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.category")
            .displayedName("Category")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add(f"{param_prefix}.feedbackCapacitor")
        managed_keys.add(f"{param_prefix}.photonEnergy")
        managed_keys.add(f"{param_prefix}.category")

        base_calcat.add_status_schema_from_enum(schema, status_prefix, LpdConstants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Feedback capacitor"] = self._get_param("feedbackCapacitor")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()
        res["Source Energy"] = self._get_param("photonEnergy")
        return res

    def category_condition(self):
        res = self.illuminated_condition()
        res["category"] = self._get_param("category")
        return res


@KARABO_CLASSINFO("LpdCorrection", deviceVersion)
class LpdCorrection(BaseCorrection):
    _correction_flag_class = CorrectionFlags
    _correction_field_names = (
        ("offset", CorrectionFlags.OFFSET),
        ("gainAmp", CorrectionFlags.GAIN_AMP),
        ("relGain", CorrectionFlags.REL_GAIN),
        ("flatfield", CorrectionFlags.FF_CORR),
        ("badPixels", CorrectionFlags.BPMASK),
    )
    _kernel_runner_class = LpdGpuRunner
    _calcat_friend_class = LpdCalcatFriend
    _constant_enum_class = LpdConstants
    _managed_keys = BaseCorrection._managed_keys.copy()

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(256)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.memoryCells")
            .setNewDefaultValue(512)
            .commit(),

            # TODO: determine ideal default
            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("frame")
            .commit(),
        )

        (
            OUTPUT_CHANNEL(expected)
            .key("preview.outputGainMap")
            .dataSchema(preview_schema)
            .commit(),
        )

        LpdCalcatFriend.add_schema(expected, LpdCorrection._managed_keys)
        add_correction_step_schema(
            expected, LpdCorrection._managed_keys, LpdCorrection._correction_field_names
        )

        # additional settings for correction steps
        (
            STRING_ELEMENT(expected)
            .key("corrections.badPixels.maskingValue")
            .displayedName("Bad pixel masking value")
            .description(
                "Any pixels masked by the bad pixel mask will have their value "
                "replaced with this. Note that this parameter is to be interpreted as "
                "a numpy.float32; use 'nan' to get NaN value."
            )
            .assignmentOptional()
            .defaultValue("nan")
            .reconfigurable()
            .commit(),
        )
        LpdCorrection._managed_keys.add("corrections.badPixels.maskingValue")

        # mandatory: manager needs this in schema
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(LpdCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.memoryCells"),
            1,
            self.unsafe_get("dataFormat.pixelsX"),
            self.unsafe_get("dataFormat.pixelsY"),
        )

    def __init__(self, config):
        super().__init__(config)
        try:
            bad_pixel_mask_value = np.float32(
                config.get("corrections.badPixels.maskingValue")
            )
        except ValueError:
            bad_pixel_mask_value = np.float32("nan")
        self._kernel_runner_init_args = {"bad_pixel_mask_value": bad_pixel_mask_value}

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_constant(constant, constant_data)
        correction_step = {
            LpdConstants.Offset: "offset",
            LpdConstants.GainAmpMap: "gainAmp",
            LpdConstants.RelativeGain: "relGain",
            LpdConstants.FFMap: "flatfield",
            LpdConstants.BadPixelsDark: "badPixels",
            LpdConstants.BadPixelsFF: "badPixels",
        }[constant]
        correction_node = f"corrections.{correction_step}"
        if not self.get(f"{correction_node}.available"):
            self.set(f"{correction_node}.available", True)
        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to GPU")

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        pulse_table = np.ravel(data_hash.get("image.pulseId"))
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            # drop the singleton dimension for kernel runner
            self.kernel_runner.load_data(image_data)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.load_cell_table(cell_table)
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        if do_generate_preview:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ) = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
                warn_func=self.log_status_warn,
            )
            (
                preview_raw,
                preview_corrected,
                preview_gain_map,
            ) = self.kernel_runner.compute_previews(
                preview_slice_index,
            )

        data_hash.set(self._image_data_path, buffer_handle)
        data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])
        data_hash.set("calngShmemPaths", [self._image_data_path])
        self._write_output(data_hash, metadata)
        if do_generate_preview:
            self._write_preview_outputs(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                    ("preview.outputGainMap", preview_gain_map),
                ),
                metadata,
            )

    def preReconfigure(self, config):
        # TODO: DRY (taken from AGIPD device)
        super().preReconfigure(config)
        if config.has("corrections.badPixels.maskingValue"):
            # only check if it is valid (let raise exception)
            # if valid, postReconfigure will use it
            np.float32(config.get("corrections.badPixels.maskingValue"))

    def postReconfigure(self):
        super().postReconfigure()
        if not hasattr(self, "_prereconfigure_update_hash"):
            return

        update = self._prereconfigure_update_hash
        if update.has("corrections.badPixels.maskingValue"):
            masking_value = np.float32(update["corrections.badPixels.maskingValue"])
            self._kernel_runner_init_args["bad_pixel_mask_value"] = masking_value
            self.kernel_runner.bad_pixel_mask_value = masking_value
